@@ -2,8 +2,8 @@
 
   dir.c -
 
-  $Author: matz $
-  $Date: 2005/07/27 02:10:56 $
+  $Author: nobu $
+  $Date: 2006/07/14 16:27:51 $
   created at: Wed Jan  5 09:51:01 JST 1994
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -67,10 +67,23 @@ char *strchr _((char*,char));
 #define lstat stat
 #endif
 
+#ifndef CASEFOLD_FILESYSTEM
+# if defined DOSISH || defined __VMS
+#   define CASEFOLD_FILESYSTEM 1
+# else
+#   define CASEFOLD_FILESYSTEM 0
+# endif
+#endif
+
 #define FNM_NOESCAPE	0x01
 #define FNM_PATHNAME	0x02
 #define FNM_DOTMATCH	0x04
 #define FNM_CASEFOLD	0x08
+#if CASEFOLD_FILESYSTEM
+#define FNM_SYSCASE	FNM_CASEFOLD
+#else
+#define FNM_SYSCASE	0
+#endif
 
 #define FNM_NOMATCH	1
 #define FNM_ERROR	2
@@ -325,7 +338,17 @@ dir_closed()
     rb_raise(rb_eIOError, "closed directory");
 }
 
+static void
+dir_check(dir)
+    VALUE dir;
+{
+    if (!OBJ_TAINTED(dir) && rb_safe_level() >= 4)
+	rb_raise(rb_eSecurityError, "Insecure: operation on untainted Dir");
+    rb_check_frozen(dir);
+}
+
 #define GetDIR(obj, dirp) do {\
+    dir_check(dir);\
     Data_Get_Struct(obj, struct dir_data, dirp);\
     if (dirp->dir == NULL) dir_closed();\
 } while (0)
@@ -536,6 +559,9 @@ dir_close(dir)
 {
     struct dir_data *dirp;
 
+    if (rb_safe_level() >= 4 && !OBJ_TAINTED(dir)) {
+	rb_raise(rb_eSecurityError, "Insecure: can't close");
+    }
     GetDIR(dir, dirp);
     closedir(dirp->dir);
     dirp->dir = NULL;
@@ -786,6 +812,17 @@ dir_s_rmdir(obj, dir)
     return INT2FIX(0);
 }
 
+static void
+sys_warning_1(mesg)
+    const char* mesg;
+{
+    rb_sys_warning("%s", mesg);
+}
+
+#define GLOB_VERBOSE	(1U << (sizeof(int) * CHAR_BIT - 1))
+#define sys_warning(val) \
+    ((flags & GLOB_VERBOSE) && rb_protect((VALUE (*)_((VALUE)))sys_warning_1, (VALUE)(val), 0))
+
 /* Return nonzero if S has any special globbing chars in it.  */
 static int
 has_magic(s, send, flags)
@@ -795,7 +832,8 @@ has_magic(s, send, flags)
     register const char *p = s;
     register char c;
     int open = 0;
-    int escape = !(flags & FNM_NOESCAPE);
+    const int escape = !(flags & FNM_NOESCAPE);
+    const int nocase = flags & FNM_CASEFOLD;
 
     while ((c = *p++) != '\0') {
 	switch (c) {
@@ -814,6 +852,11 @@ has_magic(s, send, flags)
 	  case '\\':
 	    if (escape && *p++ == '\0')
 		return Qfalse;
+	    break;
+
+	  default:
+	    if (!FNM_SYSCASE && ISALPHA(c) && nocase)
+		return Qtrue;
 	}
 
 	if (send && p >= send) break;
@@ -895,31 +938,16 @@ glob_func_caller(val)
     return Qnil;
 }
 
-static int
-glob_call_func(func, path, arg)
-    void (*func) _((const char*, VALUE));
-    const char *path;
-    VALUE arg;
-{
-    int status;
-    struct glob_args args;
+#define glob_call_func(func, path, arg) (*func)(path, arg)
 
-    args.func = func;
-    args.c = path;
-    args.v = arg;
-
-    rb_protect(glob_func_caller, (VALUE)&args, &status);
-    return status;
-}
-
-static int glob_helper _((const char *path, const char *sub, int flags, void (*func)(const char *,VALUE), VALUE arg));
+static int glob_helper _((const char *path, const char *sub, int flags, int (*func)(const char *,VALUE), VALUE arg));
 
 static int
 glob_helper(path, sub, flags, func, arg)
     const char *path;
     const char *sub;
     int flags;
-    void (*func) _((const char *, VALUE));
+    int (*func) _((const char *, VALUE));
     VALUE arg;
 {
     struct stat st;
@@ -930,13 +958,9 @@ glob_helper(path, sub, flags, func, arg)
 
     p = sub ? sub : path;
     if (!has_magic(p, 0, flags)) {
-	if (
-#if defined DOSISH
-	    TRUE
-#else
-	    !(flags & FNM_NOESCAPE)
+#if !defined DOSISH
+	if (!(flags & FNM_NOESCAPE))
 #endif
-	    )
 	{
 	    newpath = strdup(path);
 	    if (sub) {
@@ -955,7 +979,7 @@ glob_helper(path, sub, flags, func, arg)
 	else if (errno != ENOENT) {
 	    /* In case stat error is other than ENOENT and
 	       we may want to know what is wrong. */
-	    rb_protect((VALUE (*)_((VALUE)))rb_sys_warning, (VALUE)path, 0);
+	    sys_warning(path);
 	}
 	xfree(newpath);
 	return status;
@@ -982,7 +1006,7 @@ glob_helper(path, sub, flags, func, arg)
 	    magic = extract_elem(p);
 	    if (stat(dir, &st) < 0) {
 	        if (errno != ENOENT)
-		    rb_protect((VALUE (*)_((VALUE)))rb_sys_warning, (VALUE)dir, 0);
+		    sys_warning(dir);
 	        free(base);
 	        free(magic);
 	        break;
@@ -998,7 +1022,7 @@ glob_helper(path, sub, flags, func, arg)
 		}
 		dirp = opendir(dir);
 		if (dirp == NULL) {
-		    rb_protect((VALUE (*)_((VALUE)))rb_sys_warning, (VALUE)dir, 0);
+		    sys_warning(dir);
 		    free(base);
 		    free(magic);
 		    break;
@@ -1026,7 +1050,7 @@ glob_helper(path, sub, flags, func, arg)
 		    sprintf(buf, "%s%s%s", base, (BASE) ? "/" : "", dp->d_name);
 		    if (lstat(buf, &st) < 0) {
 			if (errno != ENOENT)
-			    rb_protect((VALUE (*)_((VALUE)))rb_sys_warning, (VALUE)buf, 0);
+			    sys_warning(buf);
 			continue;
 		    }
 		    if (S_ISDIR(st.st_mode)) {
@@ -1073,7 +1097,7 @@ glob_helper(path, sub, flags, func, arg)
 			    }
 			}
 			else {
-			    rb_protect((VALUE (*)_((VALUE)))rb_sys_warning, (VALUE)link->path, 0);
+			    sys_warning(link->path);
 			}
 		    }
 		    tmp = link;
@@ -1091,6 +1115,42 @@ glob_helper(path, sub, flags, func, arg)
     return status;
 }
 
+int
+ruby_glob(path, flags, func, arg)
+    const char *path;
+    int flags;
+    int (*func) _((const char *, VALUE));
+    VALUE arg;
+{
+    flags |= FNM_SYSCASE;
+    return glob_helper(path, 0, flags & ~GLOB_VERBOSE, func, arg);
+}
+
+int
+ruby_globi(path, flags, func, arg)
+    const char *path;
+    int flags;
+    int (*func) _((const char *, VALUE));
+    VALUE arg;
+{
+    return glob_helper(path, 0, flags | FNM_CASEFOLD, func, arg);
+}
+
+static int rb_glob_caller _((const char *, VALUE));
+
+static int
+rb_glob_caller(path, a)
+    const char *path;
+    VALUE a;
+{
+    int status;
+    struct glob_args *args = (struct glob_args *)a;
+
+    args->c = path;
+    rb_protect(glob_func_caller, a, &status);
+    return status;
+}
+
 static int
 rb_glob2(path, flags, func, arg)
     const char *path;
@@ -1098,23 +1158,13 @@ rb_glob2(path, flags, func, arg)
     void (*func) _((const char *, VALUE));
     VALUE arg;
 {
-    int status;
+    struct glob_args args;
 
-    status = glob_helper(path, 0, flags, func, arg);
-    return status;
-}
+    args.func = func;
+    args.v = arg;
 
-struct rb_glob_args {
-    void (*func) _((const char*, VALUE));
-    VALUE arg;
-};
-
-static void
-rb_glob_caller(path, a)
-    VALUE path, a;
-{
-    struct rb_glob_args *args = (struct rb_glob_args *)a;
-    (*args->func)(RSTRING(path)->ptr, args->arg);
+    flags |= FNM_SYSCASE;
+    return glob_helper(path, 0, flags | GLOB_VERBOSE, rb_glob_caller, (VALUE)&args);
 }
 
 void
@@ -1123,13 +1173,7 @@ rb_glob(path, func, arg)
     void (*func) _((const char*, VALUE));
     VALUE arg;
 {
-    struct rb_glob_args args;
-    int status;
-
-    args.func = func;
-    args.arg = arg;
-    status = rb_glob2(path, 0, rb_glob_caller, (VALUE)&args);
-
+    int status = rb_glob2(path, 0, func, arg);
     if (status) rb_jump_tag(status);
 }
 
@@ -1139,13 +1183,7 @@ rb_globi(path, func, arg)
     void (*func) _((const char*, VALUE));
     VALUE arg;
 {
-    struct rb_glob_args args;
-    int status;
-
-    args.func = func;
-    args.arg = arg;
-    status = rb_glob2(path, FNM_CASEFOLD, rb_glob_caller, (VALUE)&args);
-
+    int status = rb_glob2(path, FNM_CASEFOLD, func, arg);
     if (status) rb_jump_tag(status);
 }
 
@@ -1565,4 +1603,5 @@ Init_Dir()
     rb_file_const("FNM_PATHNAME", INT2FIX(FNM_PATHNAME));
     rb_file_const("FNM_DOTMATCH", INT2FIX(FNM_DOTMATCH));
     rb_file_const("FNM_CASEFOLD", INT2FIX(FNM_CASEFOLD));
+    rb_file_const("FNM_SYSCASE", INT2FIX(FNM_SYSCASE));
 }
