@@ -2,8 +2,8 @@
 
   dln.c -
 
-  $Author: matz $
-  $Date: 2005/07/19 08:25:39 $
+  $Author: nobu $
+  $Date: 2006/01/25 13:30:10 $
   created at: Tue Jan 18 17:05:06 JST 1994
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -89,8 +89,6 @@ char *getenv();
 # include <image.h>
 #endif
 
-int eaccess();
-
 #ifndef NO_DLN_LOAD
 
 #if defined(HAVE_DLOPEN) && !defined(USE_DLN_A_OUT) && !defined(_AIX) && !defined(__APPLE__) && !defined(_UNICOSMP)
@@ -1693,6 +1691,7 @@ conv_to_posix_path(win32, posix, len)
     char *p = win32;
     char *dst = posix;
 
+    posix[0] = '\0';
     for (p = win32; *p; p++)
 	if (*p == ';') {
 	    *p = 0;
