@@ -33,6 +33,64 @@ class << TclTkIp
 end
 
 
+################################################
+# use pseudo-toplevel feature of MultiTkIp ?
+if (!defined?(Use_PseudoToplevel_Feature_of_MultiTkIp) || 
+      Use_PseudoToplevel_Feature_of_MultiTkIp)
+  module MultiTkIp_PseudoToplevel_Evaluable
+    #def pseudo_toplevel_eval(body = Proc.new)
+    #  Thread.current[:TOPLEVEL] = self
+    #  begin
+    #    body.call
+    #  ensure
+    #    Thread.current[:TOPLEVEL] = nil
+    #  end
+    #end
+
+    def pseudo_toplevel_evaluable?
+      @pseudo_toplevel_evaluable
+    end
+
+    def pseudo_toplevel_evaluable=(mode)
+      @pseudo_toplevel_evaluable = (mode)? true: false
+    end
+
+    def self.extended(mod)
+      mod.__send__(:extend_object, mod)
+      mod.instance_variable_set('@pseudo_toplevel_evaluable', true)
+    end
+  end
+
+  class Object
+    alias __method_missing_alias_for_MultiTkIp__ method_missing
+    private :__method_missing_alias_for_MultiTkIp__
+
+    def method_missing(id, *args)
+      begin
+        has_top = (top = MultiTkIp.__getip.__pseudo_toplevel) && 
+          top.respond_to?(:pseudo_toplevel_evaluable?) && 
+          top.pseudo_toplevel_evaluable? && 
+          top.respond_to?(id)
+      rescue Exception => e
+        has_top = false
+      end
+
+      if has_top
+        top.__send__(id, *args)
+      else
+        __method_missing_alias_for_MultiTkIp__(id, *args)
+      end
+    end
+  end
+else
+  # dummy
+  module MultiTkIp_PseudoToplevel_Evaluable
+    def pseudo_toplevel_evaluable?
+      false
+    end
+  end
+end
+
 ################################################
 # exceptiopn to treat the return value from IP
 class MultiTkIp_OK < Exception
@@ -54,6 +112,8 @@ MultiTkIp_OK.freeze
 ################################################
 # methods for construction
 class MultiTkIp
+  BASE_DIR = File.dirname(__FILE__)
+
   @@SLAVE_IP_ID = ['slave'.freeze, '0'.taint].freeze
 
   @@IP_TABLE = {}.taint unless defined?(@@IP_TABLE)
@@ -142,7 +202,7 @@ class MultiTkIp
 
   ######################################
 
-  @@CB_ENTRY_CLASS = Class.new(TkCallbackEntry){|c|
+  @@CB_ENTRY_CLASS = Class.new(TkCallbackEntry){
     def initialize(ip, cmd)
       @ip = ip
       @cmd = cmd
@@ -158,7 +218,9 @@ class MultiTkIp
         backup_ip = current['callback_ip']
         current['callback_ip'] = @ip
         begin
-          @ip.cb_eval(@cmd, *args)
+          ret = @ip.cb_eval(@cmd, *args)
+          fail ret if ret.kind_of?(Exception)
+          ret
         rescue TkCallbackBreak, TkCallbackContinue => e
           fail e
 	rescue SecurityError => e
@@ -175,6 +237,8 @@ class MultiTkIp
             fail e
           end
         rescue Exception => e
+          fail e if e.message =~ /^TkCallback/
+
           if @ip.safe?
 	    if @ip.respond_to?(:cb_error)
               @ip.cb_error(e)
@@ -603,6 +667,7 @@ class MultiTkIp
       begin
         loop do
           sleep 1
+          receiver.kill if @interp.deleted?
           break unless receiver.alive?
         end
       rescue Exception
@@ -661,6 +726,8 @@ class MultiTkIp
     @interp = TclTkIp.new(name, _keys2opts(keys))
     @ip_name = nil
 
+    @callback_status = [].taint
+
     @system = Object.new
 
     @wait_on_mainloop = [true, 0].taint
@@ -685,6 +752,49 @@ class MultiTkIp
 
     #################################
 
+    @pseudo_toplevel = [false, nil]
+
+    def self.__pseudo_toplevel
+      Thread.current.group == ThreadGroup::Default && 
+        MultiTkIp.__getip == @@DEFAULT_MASTER &&
+        self.__pseudo_toplevel_evaluable? && @pseudo_toplevel[1]
+    end
+
+    def self.__pseudo_toplevel=(m)
+      unless (Thread.current.group == ThreadGroup::Default && 
+                MultiTkIp.__getip == @@DEFAULT_MASTER)
+        fail SecurityError, "no permission to manipulate"
+      end
+
+      # if m.kind_of?(Module) && m.respond_to?(:pseudo_toplevel_evaluable?)
+      if m.respond_to?(:pseudo_toplevel_evaluable?)
+        @pseudo_toplevel[0] = true
+        @pseudo_toplevel[1] = m
+      else
+        fail ArgumentError, 'fail to set pseudo-toplevel'
+      end
+      self
+    end
+
+    def self.__pseudo_toplevel_evaluable?
+      begin
+        @pseudo_toplevel[0] && @pseudo_toplevel[1].pseudo_toplevel_evaluable?
+      rescue Exception
+        false
+      end
+    end
+
+    def self.__pseudo_toplevel_evaluable=(mode)
+      unless (Thread.current.group == ThreadGroup::Default && 
+                MultiTkIp.__getip == @@DEFAULT_MASTER)
+        fail SecurityError, "no permission to manipulate"
+      end
+
+      @pseudo_toplevel[0] = (mode)? true: false
+    end
+
+    #################################
+
     @assign_request = Class.new(Exception){
       def self.new(target, ret)
         obj = super()
@@ -739,10 +849,40 @@ class MultiTkIp
 
     #################################
 
+    @init_ip_env_queue = Queue.new
+    Thread.new{
+      current = Thread.current
+      loop {
+        mtx, ret, table, script = @init_ip_env_queue.deq
+        begin        
+          ret[0] = table.each{|tg, ip| ip._init_ip_env(script) }
+        rescue Exception => e
+          ret[0] = e
+        ensure
+          mtx.unlock
+        end
+      }
+    }
+
+    def self.__init_ip_env__(table, script)
+      ret = []
+      mtx = Mutex.new.lock
+      @init_ip_env_queue.enq([mtx, ret, table, script])
+      mtx.lock
+      if ret[0].kind_of?(Exception)
+        raise ret[0]
+      else
+        ret[0]
+      end
+    end
+
+    #################################
+
     class << self
       undef :instance_eval
     end
   }
+
   @@DEFAULT_MASTER.freeze # defend against modification
 
   ######################################
@@ -943,6 +1083,7 @@ class MultiTkIp
     dup_keys['fill'] = :both  unless dup_keys.key?('fill')
     dup_keys['expand'] = true unless dup_keys.key?('expand')
     c = TkFrame.new(top, :container=>true).pack(dup_keys)
+    c.bind('Destroy', proc{top.destroy})
 
     # return keys
     loadTk_keys['use'] = TkWinfo.id(c)
@@ -1055,6 +1196,8 @@ class MultiTkIp
     @cb_error_proc.taint unless @cb_error_proc.tainted?
     @evloop_thread.taint unless @evloop_thread.tainted?
 
+    @callback_status = []
+
     name, safe, safe_opts, tk_opts = _parse_slaveopts(keys)
 
     safe = 4 if safe && !safe.kind_of?(Fixnum)
@@ -1105,6 +1248,8 @@ class MultiTkIp
 
     @threadgroup  = ThreadGroup.new
 
+    @pseudo_toplevel = [false, nil]
+
     @cmd_queue = Queue.new
 
 =begin
@@ -1403,7 +1548,7 @@ class MultiTkIp
   end
 
   def _init_ip_env(script)
-    script.call(self)
+    self.eval_proc{script.call(self)}
   end
 
   def _add_tk_procs(name, args, body)
@@ -1414,8 +1559,20 @@ class MultiTkIp
     }
   end
 
+  def _remove_tk_procs(*names)
+    return if slave?
+    names.each{|name|
+      name = name.to_s
+      @interp._invoke('rename', name, '')
+      @interp._invoke('interp', 'slaves').split.each{|slave|
+        @interp._invoke('interp', 'alias', slave, name, '') rescue nil
+      }
+    }
+  end
+
   def _init_ip_internal(init_ip_env, add_tk_procs)
-    init_ip_env.each{|script| script.call(self)}
+    #init_ip_env.each{|script| self.eval_proc{script.call(self)}}
+    init_ip_env.each{|script| self._init_ip_env(script)}
     add_tk_procs.each{|name, args, body| 
       if master?
         @interp._invoke('proc', name, args, body) if args && body
@@ -1439,9 +1596,21 @@ class MultiTkIp
     __getip._tk_table_list[id]
   end
   def self.create_table
-    #if __getip.slave?
-    #  raise SecurityError, "slave-IP has no permission creating a new table"
-    #end
+    if __getip.slave? 
+      begin
+        raise SecurityError, "slave-IP has no permission creating a new table"
+      rescue SecurityError => e
+        #p e.backtrace
+        # Is called on a Ruby/Tk library?
+        caller_info = e.backtrace[1]
+        if caller_info =~ %r{^#{MultiTkIp::BASE_DIR}/(tk|tkextlib)/[^:]+\.rb:}
+          # Probably, caller is a Ruby/Tk library  -->  allow creating
+        else
+          raise e
+        end
+      end
+    end
+
     id = @@TK_TABLE_LIST.size
     obj = Object.new
     @@TK_TABLE_LIST << obj
@@ -1457,15 +1626,48 @@ class MultiTkIp
 
   def self.init_ip_env(script = Proc.new)
     @@INIT_IP_ENV << script
-    @@IP_TABLE.each{|tg, ip| 
-      ip._init_ip_env(script)
-    }
+    if __getip.slave?
+      begin
+        raise SecurityError, "slave-IP has no permission initializing IP env"
+      rescue SecurityError => e
+        #p e.backtrace
+        # Is called on a Ruby/Tk library?
+        caller_info = e.backtrace[1]
+        if caller_info =~ %r{^#{MultiTkIp::BASE_DIR}/(tk|tkextlib)/[^:]+\.rb:}
+          # Probably, caller is a Ruby/Tk library  -->  allow creating
+        else
+          raise e
+        end
+      end
+    end
+
+    # @@IP_TABLE.each{|tg, ip| 
+    #   ip._init_ip_env(script)
+    # }
+    @@DEFAULT_MASTER.__init_ip_env__(@@IP_TABLE, script)
   end
 
   def self.add_tk_procs(name, args=nil, body=nil)
-    @@ADD_TK_PROCS << [name, args, body]
+    if name.kind_of?(Array) # => an array of [name, args, body]
+      name.each{|param| self.add_tk_procs(*param)}
+    else
+      name = name.to_s
+      @@ADD_TK_PROCS << [name, args, body]
+      @@IP_TABLE.each{|tg, ip| 
+        ip._add_tk_procs(name, args, body)
+      }
+    end
+  end
+
+  def self.remove_tk_procs(*names)
+    names.each{|name|
+      name = name.to_s
+      @@ADD_TK_PROCS.delete_if{|elem| 
+        elem.kind_of?(Array) && elem[0].to_s == name
+      }
+    }
     @@IP_TABLE.each{|tg, ip| 
-      ip._add_tk_procs(name, args, body)
+      ip._remove_tk_procs(*names)
     }
   end
 
@@ -1484,6 +1686,7 @@ class MultiTkIp
     @@CB_ENTRY_CLASS.new(__getip, cmd).freeze
   end
 
+=begin
   def cb_eval(cmd, *args)
     #self.eval_callback{ TkComm._get_eval_string(TkUtil.eval_cmd(cmd, *args)) }
     #ret = self.eval_callback{ TkComm._get_eval_string(TkUtil.eval_cmd(cmd, *args)) }
@@ -1496,8 +1699,105 @@ class MultiTkIp
     end
     ret
   end
+=end
+  def cb_eval(cmd, *args)
+    self.eval_callback(*args){|safe, *params|
+      $SAFE=safe if $SAFE < safe
+      # TkUtil.eval_cmd(cmd, *params)
+      TkComm._get_eval_string(TkUtil.eval_cmd(cmd, *params))
+    }
+  end
+=begin
+  def cb_eval(cmd, *args)
+    @callback_status[0] ||= TkVariable.new
+    @callback_status[1] ||= TkVariable.new
+    st, val = @callback_status
+    th = Thread.new{
+      self.eval_callback(*args){|safe, *params|
+        #p [status, val, safe, *params]
+        $SAFE=safe if $SAFE < safe
+        begin
+          TkComm._get_eval_string(TkUtil.eval_cmd(cmd, *params))
+        rescue TkCallbackContinue
+          st.value = 4
+        rescue TkCallbackBreak
+          st.value = 3
+        rescue TkCallbackReturn
+          st.value = 2
+        rescue Exception => e
+          val.value = e.message
+          st.value = 1
+        else
+          st.value = 0
+        end
+      }
+    }
+    begin
+      st.wait
+      status = st.numeric
+      retval = val.value
+    rescue => e
+      fail e
+    end
+
+    if status == 1
+      fail RuntimeError, retval
+    elsif status == 2
+      fail TkCallbackReturn, "Tk callback returns 'return' status"
+    elsif status == 3
+      fail TkCallbackBreak, "Tk callback returns 'break' status"
+    elsif status == 4
+      fail TkCallbackContinue, "Tk callback returns 'continue' status"
+    else
+      ''
+    end
+  end
+=end
+
 end
 
+# pseudo-toplevel operation support
+class MultiTkIp
+  # instance method
+  def __pseudo_toplevel
+    ip = MultiTkIp.__getip
+    (ip == @@DEFAULT_MASTER || ip == self) &&
+      self.__pseudo_toplevel_evaluable? && @pseudo_toplevel[1]
+  end
+
+  def __pseudo_toplevel=(m)
+    unless (Thread.current.group == ThreadGroup::Default && 
+              MultiTkIp.__getip == @@DEFAULT_MASTER)
+      fail SecurityError, "no permission to manipulate"
+    end
+
+    # if m.kind_of?(Module) && m.respond_to?(:pseudo_toplevel_evaluable?)
+    if m.respond_to?(:pseudo_toplevel_evaluable?)
+      @pseudo_toplevel[0] = true
+      @pseudo_toplevel[1] = m
+    else
+      fail ArgumentError, 'fail to set pseudo-toplevel'
+    end
+    self
+  end
+
+  def __pseudo_toplevel_evaluable?
+    begin
+      @pseudo_toplevel[0] && @pseudo_toplevel[1].pseudo_toplevel_evaluable?
+    rescue Exception
+      false
+    end
+  end
+
+  def __pseudo_toplevel_evaluable=(mode)
+    unless (Thread.current.group == ThreadGroup::Default && 
+              MultiTkIp.__getip == @@DEFAULT_MASTER)
+      fail SecurityError, "no permission to manipulate"
+    end
+
+    @pseudo_toplevel[0] = (mode)? true: false
+  end
+end
 
 # evaluate a procedure on the proper interpreter
 class MultiTkIp
@@ -1701,6 +2001,35 @@ class MultiTkIp
   alias background_eval_string bg_eval_string
   alias bg_eval_str bg_eval_string
   alias background_eval_str bg_eval_string
+
+  def eval(*args, &blk)
+    if block_given?
+      eval_proc(*args, &blk)
+    elsif args[0]
+      if args[0].respond_to?(:call)
+        eval_proc(*args)
+      else
+        eval_string(*args)
+      end
+    else
+      fail ArgumentError, "no argument to eval"
+    end
+  end
+
+  def bg_eval(*args, &blk)
+    if block_given?
+      bg_eval_proc(*args, &blk)
+    elsif args[0]
+      if args[0].respond_to?(:call)
+        bg_eval_proc(*args)
+      else
+        bg_eval_string(*args)
+      end
+    else
+      fail ArgumentError, "no argument to eval"
+    end
+  end
+  alias background_eval bg_eval
 end
 
 class << MultiTkIp
@@ -1710,7 +2039,39 @@ class << MultiTkIp
     __getip.eval_proc(*args, &blk)
   end
   alias call eval_proc
-  alias eval_string eval_proc
+
+  def bg_eval_proc(*args, &blk)
+    # class ==> interp object
+    __getip.bg_eval_proc(*args, &blk)
+  end
+  alias background_eval_proc bg_eval_proc
+  alias thread_eval_proc bg_eval_proc
+  alias bg_call bg_eval_proc
+  alias background_call bg_eval_proc
+
+  def eval_string(cmd, *eval_args)
+    # class ==> interp object
+    __getip.eval_string(cmd, *eval_args)
+  end
+  alias eval_str eval_string
+
+  def bg_eval_string(cmd, *eval_args)
+    # class ==> interp object
+    __getip.bg_eval_string(cmd, *eval_args)
+  end
+  alias background_eval_string bg_eval_string
+  alias bg_eval_str bg_eval_string
+  alias background_eval_str bg_eval_string
+
+  def eval(*args, &blk)
+    # class ==> interp object
+    __getip.eval(*args, &blk)
+  end
+  def bg_eval(*args, &blk)
+    # class ==> interp object
+    __getip.bg_eval(*args, &blk)
+  end
+  alias background_eval bg_eval
 end
 
 
@@ -1905,6 +2266,10 @@ class << MultiTkIp
     __getip._unset_global_var2(var, idx)
   end
 
+  def _make_menu_embeddable(menu_path)
+    __getip._make_menu_embeddable(menu_path)
+  end
+
   def _split_tklist(str)
     __getip._split_tklist(str)
   end
@@ -2346,6 +2711,11 @@ class MultiTkIp
     @interp._unset_global_var2(var, idx)
   end
 
+  def _make_menu_embeddable(menu_path)
+    raise SecurityError, "no permission to manipulate" unless self.manipulable?
+    @interp._make_menu_embeddable(menu_path)
+  end
+
   def _split_tklist(str)
     raise SecurityError, "no permission to manipulate" unless self.manipulable?
     @interp._split_tklist(str)
