@@ -8,8 +8,13 @@ require 'shellwords'
 CONFIG = Config::MAKEFILE_CONFIG
 ORIG_LIBPATH = ENV['LIB']
 
-SRC_EXT = ["c", "cc", "m", "cxx", "cpp", "C"]
+CXX_EXT = %w[cc cxx cpp]
+if /mswin|bccwin|mingw|msdosdjgpp|human|os2/ !~ CONFIG['build_os']
+  CXX_EXT.concat(%w[C])
+end
+SRC_EXT = %w[c m] << CXX_EXT
 $static = $config_h = nil
+$default_static = $static
 
 unless defined? $configure_args
   $configure_args = {}
@@ -20,6 +25,7 @@ unless defined? $configure_args
   for arg in Shellwords::shellwords(args)
     arg, val = arg.split('=', 2)
     next unless arg
+    arg.tr!('_', '-')
     if arg.sub!(/^(?!--)/, '--')
       val or next
       arg.downcase!
@@ -30,6 +36,7 @@ unless defined? $configure_args
   for arg in ARGV
     arg, val = arg.split('=', 2)
     next unless arg
+    arg.tr!('_', '-')
     if arg.sub!(/^(?!--)/, '--')
       val or next
       arg.downcase!
@@ -38,7 +45,6 @@ unless defined? $configure_args
   end
 end
 
-$srcdir = CONFIG["srcdir"]
 $libdir = CONFIG["libdir"]
 $rubylibdir = CONFIG["rubylibdir"]
 $archdir = CONFIG["archdir"]
@@ -53,6 +59,8 @@ $cygwin = /cygwin/ =~ RUBY_PLATFORM
 $human = /human/ =~ RUBY_PLATFORM
 $netbsd = /netbsd/ =~ RUBY_PLATFORM
 $os2 = /os2/ =~ RUBY_PLATFORM
+$beos = /beos/ =~ RUBY_PLATFORM
+$solaris = /solaris/ =~ RUBY_PLATFORM
 
 def config_string(key, config = CONFIG)
   s = config[key] and !s.empty? and block_given? ? yield(s) : s
@@ -71,18 +79,45 @@ INSTALL_DIRS = [
   [dir_re('sitearchdir'), "$(RUBYARCHDIR)"]
 ]
 
+def install_dirs(target_prefix = nil)
+  if $extout
+    dirs = [
+      ['RUBYCOMMONDIR', '$(extout)'],
+      ['RUBYLIBDIR',    '$(extout)$(target_prefix)'],
+      ['RUBYARCHDIR',   '$(extout)/$(arch)$(target_prefix)'],
+      ['extout',        "#$extout"],
+      ['extout_prefix', "#$extout_prefix"],
+    ]
+  elsif $extmk
+    dirs = [
+      ['RUBYCOMMONDIR', '$(rubylibdir)'],
+      ['RUBYLIBDIR',    '$(rubylibdir)$(target_prefix)'],
+      ['RUBYARCHDIR',   '$(archdir)$(target_prefix)'],
+    ]
+  else
+    dirs = [
+      ['RUBYCOMMONDIR', '$(sitedir)$(target_prefix)'],
+      ['RUBYLIBDIR',    '$(sitelibdir)$(target_prefix)'],
+      ['RUBYARCHDIR',   '$(sitearchdir)$(target_prefix)'],
+    ]
+  end
+  dirs << ['target_prefix', (target_prefix ? "/#{target_prefix}" : "")]
+  dirs
+end
+
 def map_dir(dir, map = nil)
   map ||= INSTALL_DIRS
   map.inject(dir) {|dir, (orig, new)| dir.gsub(orig, new)}
 end
 
-libdir = File.dirname(__FILE__)
-$extmk = libdir != Config::CONFIG["rubylibdir"]
-if not $extmk and File.exist? Config::CONFIG["archdir"] + "/ruby.h"
+topdir = File.dirname(libdir = File.dirname(__FILE__))
+extdir = File.expand_path("ext", topdir)
+$extmk = File.expand_path($0)[0, extdir.size+1] == extdir+"/"
+if not $extmk and File.exist?(Config::CONFIG["archdir"] + "/ruby.h")
   $hdrdir = $topdir = Config::CONFIG["archdir"]
-elsif File.exist? $srcdir + "/ruby.h"
-  $topdir = Config::CONFIG["compile_dir"]
-  $hdrdir = $srcdir
+elsif File.exist?(($top_srcdir ||= topdir)  + "/ruby.h") and
+    File.exist?(($topdir ||= Config::CONFIG["topdir"]) + "/config.h")
+  $hdrdir = $top_srcdir
 else
   abort "can't find header files for ruby."
 end
@@ -134,6 +169,7 @@ module Logging
   @logfile = 'mkmf.log'
   @orgerr = $stderr.dup
   @orgout = $stdout.dup
+  @postpone = 0
 
   def self::open
     @log ||= File::open(@logfile, 'w')
@@ -162,7 +198,7 @@ module Logging
   end
   
   def self::postpone
-    tmplog = "mkmftmp.log"
+    tmplog = "mkmftmp#{@postpone += 1}.log"
     open do
       log, *save = @log, @logfile, @orgout, @orgerr
       @log, @logfile, @orgout, @orgerr = nil, tmplog, log, log
@@ -172,6 +208,7 @@ module Logging
         File::open(tmplog) {|t| FileUtils.copy_stream(t, log)}
       ensure
         @log, @logfile, @orgout, @orgerr = log, *save
+        @postpone -= 1
         rm_f tmplog
       end
     end
@@ -179,7 +216,6 @@ module Logging
 end
 
 def xsystem command
-  Config.expand(command)
   Logging::open do
     puts command.quote
     system(command)
@@ -187,7 +223,6 @@ def xsystem command
 end
 
 def xpopen command, *mode, &block
-  Config.expand(command)
   Logging::open do
     case mode[0]
     when nil, /^r/
@@ -200,17 +235,22 @@ def xpopen command, *mode, &block
 end
 
 def log_src(src)
-  Logging::message <<"EOM", src
+  src = src.split(/^/)
+  fmt = "%#{src.size.to_s.size}d: %s"
+  Logging::message <<"EOM"
 checked program was:
 /* begin */
-%s/* end */
+EOM
+  src.each_with_index {|line, no| Logging::message fmt, no+1, line}
+  Logging::message <<"EOM"
+/* end */
 
 EOM
 end
 
 def create_tmpsrc(src)
   src = yield(src) if block_given?
-  src = src.sub(/[^\n]\z/, "\\&\n")
+  src = src.gsub(/[ \t]+$/, '').gsub(/\A\n+|^\n+$/, '').sub(/[^\n]\z/, "\\&\n")
   open(CONFTEST_C, "wb") do |cfile|
     cfile.print src
   end
@@ -226,31 +266,31 @@ end
 
 def link_command(ldflags, opt="", libpath=$LIBPATH)
   Config::expand(TRY_LINK.dup,
-		 'hdrdir' => $hdrdir,
-		 'src' => CONFTEST_C,
-		 'INCFLAGS' => $INCFLAGS,
-		 'CPPFLAGS' => $CPPFLAGS,
-		 'CFLAGS' => "#$CFLAGS",
-		 'ARCH_FLAG' => "#$ARCH_FLAG",
-		 'LDFLAGS' => "#$LDFLAGS #{ldflags}",
-		 'LIBPATH' => libpathflag(libpath),
-		 'LOCAL_LIBS' => "#$LOCAL_LIBS #$libs",
-		 'LIBS' => "#$LIBRUBYARG_STATIC #{opt} #$LIBS")
+                 CONFIG.merge('hdrdir' => $hdrdir.quote,
+                              'src' => CONFTEST_C,
+                              'INCFLAGS' => $INCFLAGS,
+                              'CPPFLAGS' => $CPPFLAGS,
+                              'CFLAGS' => "#$CFLAGS",
+                              'ARCH_FLAG' => "#$ARCH_FLAG",
+                              'LDFLAGS' => "#$LDFLAGS #{ldflags}",
+                              'LIBPATH' => libpathflag(libpath),
+                              'LOCAL_LIBS' => "#$LOCAL_LIBS #$libs",
+                              'LIBS' => "#$LIBRUBYARG_STATIC #{opt} #$LIBS"))
 end
 
 def cc_command(opt="")
-  "$(CC) -c #$INCFLAGS -I#{$hdrdir} " \
-  "#$CPPFLAGS #$CFLAGS #$ARCH_FLAG #{opt} #{CONFTEST_C}"
+  Config::expand("$(CC) #$INCFLAGS #$CPPFLAGS #$CFLAGS #$ARCH_FLAG #{opt} -c #{CONFTEST_C}",
+		 CONFIG.merge('hdrdir' => $hdrdir.quote, 'srcdir' => $srcdir.quote))
 end
 
 def cpp_command(outfile, opt="")
-  "$(CPP) #$INCFLAGS -I#{$hdrdir} " \
-  "#$CPPFLAGS #$CFLAGS #{opt} #{CONFTEST_C} #{outfile}"
+  Config::expand("$(CPP) #$INCFLAGS #$CPPFLAGS #$CFLAGS #{opt} #{CONFTEST_C} #{outfile}",
+		 CONFIG.merge('hdrdir' => $hdrdir.quote, 'srcdir' => $srcdir.quote))
 end
 
 def libpathflag(libpath=$LIBPATH)
   libpath.map{|x|
-    (x == "$(topdir)" ? LIBPATHFLAG : LIBPATHFLAG+RPATHFLAG) % x
+    (x == "$(topdir)" ? LIBPATHFLAG : LIBPATHFLAG+RPATHFLAG) % x.quote
   }.join
 end
 
@@ -285,54 +325,78 @@ def cpp_include(header)
   end
 end
 
+def with_cppflags(flags)
+  cppflags = $CPPFLAGS
+  $CPPFLAGS = flags
+  ret = yield
+ensure
+  $CPPFLAGS = cppflags unless ret
+end
+
+def with_cflags(flags)
+  cflags = $CFLAGS
+  $CFLAGS = flags
+  ret = yield
+ensure
+  $CFLAGS = cflags unless ret
+end
+
+def with_ldflags(flags)
+  ldflags = $LDFLAGS
+  $LDFLAGS = flags
+  ret = yield
+ensure
+  $LDFLAGS = ldflags unless ret
+end
+
 def try_static_assert(expr, headers = nil, opt = "", &b)
   headers = cpp_include(headers)
   try_compile(<<SRC, opt, &b)
 #{COMMON_HEADERS}
 #{headers}
 /*top*/
-int tmp[(#{expr}) ? 1 : -1];
+int conftest_const[(#{expr}) ? 1 : -1];
 SRC
 end
 
 def try_constant(const, headers = nil, opt = "", &b)
-  headers = cpp_include(headers)
+  includes = cpp_include(headers)
   if CROSS_COMPILING
-    unless try_compile(<<"SRC", opt, &b)
-#{COMMON_HEADERS}
-#{headers}
-/*top*/
-int tmp = #{const};
-SRC
-      return nil
-    end
-    if try_static_assert("#{const} < 0", headers, opt)
+    if try_static_assert("#{const} > 0", headers, opt)
+      # positive constant
+    elsif try_static_assert("#{const} < 0", headers, opt)
       neg = true
       const = "-(#{const})"
     elsif try_static_assert("#{const} == 0", headers, opt)
       return 0
+    else
+      # not a constant
+      return nil
     end
     upper = 1
-    until try_static_assert("#{const} < #{upper}", headers, opt)
+    lower = 0
+    until try_static_assert("#{const} <= #{upper}", headers, opt)
       lower = upper
       upper <<= 1
     end
     return nil unless lower
-    until try_static_assert("#{const} == #{upper}", headers, opt)
-      if try_static_assert("#{const} > #{(upper+lower)/2}", headers, opt)
-        lower = (upper+lower)/2
+    while upper > lower + 1
+      mid = (upper + lower) / 2
+      if try_static_assert("#{const} > #{mid}", headers, opt)
+        lower = mid
       else
-        upper = (upper+lower)/2
+        upper = mid
       end
     end
     upper = -upper if neg
     return upper
   else
     src = %{#{COMMON_HEADERS}
-#{headers}
+#{includes}
 #include <stdio.h>
 /*top*/
-int main() {printf("%d\\n", (int)(#{const})); return 0;}
+int conftest_const = (int)(#{const});
+int main() {printf("%d\\n", conftest_const); return 0;}
 }
     if try_link0(src, opt, &b)
       xpopen("./conftest") do |f|
@@ -359,6 +423,17 @@ int t() { void ((*volatile p)()); p = (void ((*)()))#{func}; return 0; }
 SRC
 end
 
+def try_var(var, headers = nil, &b)
+  headers = cpp_include(headers)
+  try_compile(<<"SRC", &b)
+#{COMMON_HEADERS}
+#{headers}
+/*top*/
+int main() { return 0; }
+int t() { const volatile void *volatile p; p = (void *)&#{var}; return 0; }
+SRC
+end
+
 def egrep_cpp(pat, src, opt = "", &b)
   src = create_tmpsrc(src, &b)
   xpopen(cpp_command('', opt)) do |f|
@@ -387,10 +462,11 @@ end
 
 def macro_defined?(macro, src, opt = "", &b)
   src = src.sub(/[^\n]\z/, "\\&\n")
-  try_cpp(src + <<"SRC", opt, &b)
+  try_compile(src + <<"SRC", opt, &b)
 /*top*/
 #ifndef #{macro}
 # error
+>>>>>> #{macro} undefined <<<<<<
 #endif
 SRC
 end
@@ -414,7 +490,7 @@ def install_files(mfile, ifiles, map = nil, srcprefix = nil)
   ifiles.each do |files, dir, prefix|
     dir = map_dir(dir, map)
     prefix = %r|\A#{Regexp.quote(prefix)}/?| if prefix
-    if( files[0,2] == "./" )
+    if /\A\.\// =~ files
       # install files which are in current working directory.
       files = files[2..-1]
       len = nil
@@ -457,14 +533,14 @@ def message(*s)
   end
 end
 
-def checking_for(m)
+def checking_for(m, fmt = nil)
   f = caller[0][/in `(.*)'$/, 1] and f << ": " #` for vim
-  m = "checking for #{m}... "
+  m = "checking #{'for ' if /\Acheck/ !~ f}#{m}... "
   message "%s", m
   a = r = nil
   Logging::postpone do
     r = yield
-    a = r ? "yes\n" : "no\n"
+    a = (fmt ? fmt % r : r ? "yes" : "no") << "\n"
     "#{f}#{m}-------------------- #{a}\n"
   end
   message(a)
@@ -472,7 +548,15 @@ def checking_for(m)
   r
 end
 
-def have_library(lib, func = nil, &b)
+def have_macro(macro, headers = nil, opt = "", &b)
+  m = "#{macro}"
+  m << " in #{headers.inspect}" if headers
+  checking_for m do
+    macro_defined?(macro, cpp_include(headers), opt, &b)
+  end
+end
+
+def have_library(lib, func = nil, headers = nil, &b)
   func = "main" if !func or func.empty?
   lib = with_config(lib+'lib', lib)
   checking_for "#{func}() in #{LIBARG%lib}" do
@@ -480,7 +564,7 @@ def have_library(lib, func = nil, &b)
       true
     else
       libs = append_library($libs, lib)
-      if try_func(func, libs, &b)
+      if try_func(func, libs, headers, &b)
         $libs = libs
         true
       else
@@ -493,6 +577,7 @@ end
 def find_library(lib, func, *paths, &b)
   func = "main" if !func or func.empty?
   lib = with_config(lib+'lib', lib)
+  paths = paths.collect {|path| path.split(File::PATH_SEPARATOR)}.flatten
   checking_for "#{func}() in #{LIBARG%lib}" do
     libpath = $LIBPATH
     libs = append_library($libs, lib)
@@ -522,6 +607,17 @@ def have_func(func, headers = nil, &b)
   end
 end
 
+def have_var(var, headers = nil, &b)
+  checking_for "#{var}" do
+    if try_var(var, headers, &b)
+      $defs.push(format("-DHAVE_%s", var.upcase))
+      true
+    else
+      false
+    end
+  end
+end
+
 def have_header(header, &b)
   checking_for header do
     if try_cpp(cpp_include(header), &b)
@@ -533,6 +629,26 @@ def have_header(header, &b)
   end
 end
 
+def find_header(header, *paths)
+  header = cpp_include(header)
+  checking_for header do
+    if try_cpp(header)
+      true
+    else
+      found = false
+      paths.each do |dir|
+        opt = "-I#{dir}".quote
+        if try_cpp(header, opt)
+          $INCFLAGS << " " << opt
+          found = true
+          break
+        end
+      end
+      found
+    end
+  end
+end
+
 def have_struct_member(type, member, header = nil, &b)
   checking_for "#{type}.#{member}" do
     if try_compile(<<"SRC", &b)
@@ -550,17 +666,17 @@ SRC
   end
 end
 
-def have_type(type, header = nil, opt = "", &b)
+def have_type(type, headers = nil, opt = "", &b)
   checking_for type do
-    header = cpp_include(header)
+    headers = cpp_include(headers)
     if try_compile(<<"SRC", opt, &b) or (/\A\w+\z/n =~ type && try_compile(<<"SRC", opt, &b))
 #{COMMON_HEADERS}
-#{header}
+#{headers}
 /*top*/
 static #{type} t;
 SRC
 #{COMMON_HEADERS}
-#{header}
+#{headers}
 /*top*/
 static #{type} *t;
 SRC
@@ -574,20 +690,85 @@ end
 
 def check_sizeof(type, header = nil, &b)
   expr = "sizeof(#{type})"
-  m = "checking size of #{type}... "
-  message "%s", m
-  Logging::message "check_sizeof: %s--------------------\n", m
-  if size = try_constant(expr, header, &b)
-    $defs.push(format("-DSIZEOF_%s=%d", type.upcase.tr_s("^A-Z0-9_", "_"), size))
+  fmt = "%d"
+  def fmt.%(x)
+    x ? super : "failed"
+  end
+  checking_for("size of #{type}", fmt) do
+    if size = try_constant(expr, header, &b)
+      $defs.push(format("-DSIZEOF_%s=%d", type.upcase.tr_s("^A-Z0-9_", "_"), size))
+      size
+    end
+  end
+end
+
+def scalar_ptr_type?(type, member = nil, headers = nil, &b)
+  try_compile(<<"SRC", &b)   # pointer
+#{COMMON_HEADERS}
+#{cpp_include(headers)}
+/*top*/
+volatile #{type} conftestval;
+int main() { return 0; }
+int t() {return (int)(1-*(conftestval#{member ? ".#{member}" : ""}));}
+SRC
+end
+
+def scalar_type?(type, member = nil, headers = nil, &b)
+  try_compile(<<"SRC", &b)   # pointer
+#{COMMON_HEADERS}
+#{cpp_include(headers)}
+/*top*/
+volatile #{type} conftestval;
+int main() { return 0; }
+int t() {return (int)(1-(conftestval#{member ? ".#{member}" : ""}));}
+SRC
+end
+
+def what_type?(type, member = nil, headers = nil, &b)
+  m = "#{type}"
+  name = type
+  if member
+    m << "." << member
+    name = "(((#{type} *)0)->#{member})"
+  end
+  m << " in #{headers.inspect}" if headers
+  fmt = "seems %s"
+  def fmt.%(x)
+    x ? super : "unknown"
+  end
+  checking_for m, fmt do
+    if scalar_ptr_type?(type, member, headers, &b)
+      if try_static_assert("sizeof(*#{name}) == 1", headers)
+        "string"
+      end
+    elsif scalar_type?(type, member, headers, &b)
+      if try_static_assert("sizeof(#{name}) > sizeof(long)", headers)
+        "long long"
+      elsif try_static_assert("sizeof(#{name}) > sizeof(int)", headers)
+        "long"
+      elsif try_static_assert("sizeof(#{name}) > sizeof(short)", headers)
+        "int"
+      elsif try_static_assert("sizeof(#{name}) > 1", headers)
+        "short"
+      else
+        "char"
+      end
+    end
   end
-  message(a = size ? "#{size}\n" : "failed\n")
-  Logging::message "-------------------- %s\n", a
-  size
 end
 
 def find_executable0(bin, path = nil)
-  path = (path || ENV['PATH']).split(File::PATH_SEPARATOR)
   ext = config_string('EXEEXT')
+  if File.expand_path(bin) == bin
+    return bin if File.executable?(bin)
+    return file if ext and File.executable?(file = bin + ext)
+    return nil
+  end
+  if path ||= ENV['PATH']
+    path = path.split(File::PATH_SEPARATOR)
+  else
+    path = %w[/usr/local/bin /usr/ucb /usr/bin /bin]
+  end
   file = nil
   path.each do |dir|
     return file if File.executable?(file = File.join(dir, bin))
@@ -602,57 +783,79 @@ def find_executable(bin, path = nil)
   end
 end
 
-def arg_config(config, default=nil)
-  $configure_args.fetch(config, default)
+def arg_config(config, *defaults, &block)
+  $arg_config << [config, *defaults]
+  defaults << nil if !block and defaults.empty?
+  $configure_args.fetch(config.tr('_', '-'), *defaults, &block)
 end
 
-def with_config(config, default=nil)
-  unless /^--with-/ =~ config
-    config = '--with-' + config
+def with_config(config, *defaults)
+  config = config.sub(/^--with[-_]/, '')
+  val = arg_config("--with-"+config) do
+    if arg_config("--without-"+config)
+      false
+    elsif block_given?
+      yield(config, *defaults)
+    else
+      break *defaults
+    end
+  end
+  case val
+  when "yes"
+    true
+  when "no"
+    false
+  else
+    val
   end
-  arg_config(config, default)
 end
 
-def enable_config(config, default=nil)
+def enable_config(config, *defaults)
   if arg_config("--enable-"+config)
     true
   elsif arg_config("--disable-"+config)
     false
+  elsif block_given?
+    yield(config, *defaults)
   else
-    default
+    return *defaults
   end
 end
 
 def create_header(header = "extconf.h")
   message "creating %s\n", header
-  if $defs.length > 0
-    sym = header.tr("a-z./\055", "A-Z___")
-    open(header, "w") do |hfile|
-      hfile.print "#ifndef #{sym}\n#define #{sym}\n"
-      for line in $defs
-	case line
-	when /^-D([^=]+)(?:=(.*))?/
-	  hfile.print "#define #$1 #{$2 || 1}\n"
-	when /^-U(.*)/
-	  hfile.print "#undef #$1\n"
-	end
-      end
-      hfile.print "#endif\n"
+  sym = header.tr("a-z./\055", "A-Z___")
+  hdr = ["#ifndef #{sym}\n#define #{sym}\n"]
+  for line in $defs
+    case line
+    when /^-D([^=]+)(?:=(.*))?/
+      hdr << "#define #$1 #{$2 ? Shellwords.shellwords($2)[0] : 1}\n"
+    when /^-U(.*)/
+      hdr << "#undef #$1\n"
     end
   end
+  hdr << "#endif\n"
+  hdr = hdr.join
+  unless (IO.read(header) == hdr rescue false)
+    open(header, "w") do |hfile|
+      hfile.write(hdr)
+    end
+  end
+  $extconf_h = header
 end
 
 def dir_config(target, idefault=nil, ldefault=nil)
   if dir = with_config(target + "-dir", (idefault unless ldefault))
-    defaults = dir.split(File::PATH_SEPARATOR)
+    defaults = Array === dir ? dir : dir.split(File::PATH_SEPARATOR)
     idefault = ldefault = nil
   end
 
   idir = with_config(target + "-include", idefault)
+  $arg_config.last[1] ||= "${#{target}-dir}/include"
   ldir = with_config(target + "-lib", ldefault)
+  $arg_config.last[1] ||= "${#{target}-dir}/lib"
 
-#  idirs = idir ? idir.split(File::PATH_SEPARATOR) : []
-  idirs = idir.split(File::PATH_SEPARATOR) rescue []
+  idirs = idir ? Array === idir ? idir : idir.split(File::PATH_SEPARATOR) : []
   if defaults
     idirs.concat(defaults.collect {|dir| dir + "/include"})
     idir = ([idir] + idirs).compact.join(File::PATH_SEPARATOR)
@@ -665,7 +868,7 @@ def dir_config(target, idefault=nil, ldefault=nil)
     end
   end
 
-  ldirs = ldir ? ldir.split(File::PATH_SEPARATOR) : []
+  ldirs = ldir ? Array === ldir ? ldir : ldir.split(File::PATH_SEPARATOR) : []
   if defaults
     ldirs.concat(defaults.collect {|dir| dir + "/lib"})
     ldir = ([ldir] + ldirs).compact.join(File::PATH_SEPARATOR)
@@ -676,16 +879,23 @@ def dir_config(target, idefault=nil, ldefault=nil)
 end
 
 def pkg_config(pkg)
-  unless defined?($PKGCONFIG)
-    if pkgconfig = with_config("pkg-config", !CROSS_COMPILING && "pkg-config")
-      find_executable0(pkgconfig) or pkgconfig = nil
-    end
-    $PKGCONFIG = pkgconfig
+  if pkgconfig = with_config("#{pkg}-config") and find_executable0(pkgconfig)
+    # iff package specific config command is given
+    get = proc {|opt| `#{pkgconfig} --#{opt}`.chomp}
+  elsif ($PKGCONFIG ||= 
+         (pkgconfig = with_config("pkg-config", ("pkg-config" unless CROSS_COMPILING))) &&
+         find_executable0(pkgconfig) && pkgconfig) and
+      system("#{$PKGCONFIG} --exists #{pkg}")
+    # default to pkg-config command
+    get = proc {|opt| `#{$PKGCONFIG} --#{opt} #{pkg}`.chomp}
+  elsif find_executable0(pkgconfig = "#{pkg}-config")
+    # default to package specific config command, as a last resort.
+    get = proc {|opt| `#{pkgconfig} --#{opt}`.chomp}
   end
-  if $PKGCONFIG and system("#{$PKGCONFIG} --exists #{pkg}")
-    cflags = `#{$PKGCONFIG} --cflags #{pkg}`.chomp
-    ldflags = `#{$PKGCONFIG} --libs #{pkg}`.chomp
-    libs = `#{$PKGCONFIG} --libs-only-l #{pkg}`.chomp
+  if get
+    cflags = get['cflags']
+    ldflags = get['libs']
+    libs = get['libs-only-l']
     ldflags = (Shellwords.shellwords(ldflags) - Shellwords.shellwords(libs)).quote.join(" ")
     $CFLAGS += " " << cflags
     $LDFLAGS += " " << ldflags
@@ -694,6 +904,9 @@ def pkg_config(pkg)
     Logging::message "cflags: %s\nldflags: %s\nlibs: %s\n\n",
                      cflags, ldflags, libs
     [cflags, ldflags, libs]
+  else
+    Logging::message "package configuration for %s is not found\n", pkg
+    nil
   end
 end
 
@@ -707,15 +920,26 @@ end
 
 def configuration(srcdir)
   mk = []
+  vpath = %w[$(srcdir) $(topdir) $(hdrdir)]
+  if !CROSS_COMPILING
+    case CONFIG['build_os']
+    when 'cygwin'
+      if CONFIG['target_os'] != 'cygwin'
+        vpath.each {|p| p.sub!(/.*/, '$(shell cygpath -u \&)')}
+      end
+    when 'msdosdjgpp', 'mingw32'
+      CONFIG['PATH_SEPARATOR'] = ';'
+    end
+  end
   mk << %{
 SHELL = /bin/sh
 
 #### Start of system configuration section. ####
 
-srcdir = #{srcdir}
-topdir = #{$topdir}
-hdrdir = #{$extmk ? $hdrdir : '$(topdir)'}
-VPATH = #{$mingw && CONFIG['build_os'] == 'cygwin' ? '$(shell cygpath -u $(srcdir))' : '$(srcdir)'}
+srcdir = #{srcdir.gsub(/\$\((srcdir)\)|\$\{(srcdir)\}/) {CONFIG[$1||$2]}.quote}
+topdir = #{($extmk ? CONFIG["topdir"] : $topdir).quote}
+hdrdir = #{$extmk ? CONFIG["hdrdir"].quote : '$(topdir)'}
+VPATH = #{vpath.join(CONFIG['PATH_SEPARATOR'])}
 }
   drive = File::PATH_SEPARATOR == ';' ? /\A\w:/ : /\A/
   if destdir = CONFIG["prefix"].scan(drive)[0] and !destdir.empty?
@@ -727,9 +951,16 @@ VPATH = #{$mingw && CONFIG['build_os'] == 'cygwin' ? '$(shell cygpath -u $(srcdi
   end
   CONFIG.each do |key, var|
     next if /^abs_/ =~ key
-    next unless /^(?:src|top|(.*))dir$/ =~ key and $1
+    next unless /^(?:src|top|hdr|(.*))dir$/ =~ key and $1
     mk << "#{key} = #{with_destdir(var.sub(drive, ''))}\n"
   end
+  if !$extmk and !$configure_args.has_key?('--ruby') and
+      sep = config_string('BUILD_FILE_SEPARATOR')
+    sep = ":/=#{sep}"
+  else
+    sep = ""
+  end
+  extconf_h = $extconf_h ? "-DRUBY_EXTCONF_H=\\\"$(RUBY_EXTCONF_H)\\\" " : $defs.join(" ")<<" "
   mk << %{
 CC = #{CONFIG['CC']}
 LIBRUBY = #{CONFIG['LIBRUBY']}
@@ -737,8 +968,10 @@ LIBRUBY_A = #{CONFIG['LIBRUBY_A']}
 LIBRUBYARG_SHARED = #$LIBRUBYARG_SHARED
 LIBRUBYARG_STATIC = #$LIBRUBYARG_STATIC
 
+RUBY_EXTCONF_H = #{$extconf_h}
 CFLAGS   = #{CONFIG['CCDLFLAGS'] unless $static} #$CFLAGS #$ARCH_FLAG
-CPPFLAGS = -I. -I$(topdir) -I$(hdrdir) -I$(srcdir) #{$defs.join(" ")} #{$CPPFLAGS}
+INCFLAGS = -I. #$INCFLAGS
+CPPFLAGS = #{extconf_h}#{$CPPFLAGS}
 CXXFLAGS = $(CFLAGS) #{CONFIG['CXXFLAGS']}
 DLDFLAGS = #$LDFLAGS #$DLDFLAGS #$ARCH_FLAG
 LDSHARED = #{CONFIG['LDSHARED']}
@@ -751,14 +984,17 @@ arch = #{CONFIG['arch']}
 sitearch = #{CONFIG['sitearch']}
 ruby_version = #{Config::CONFIG['ruby_version']}
 ruby = #{$ruby}
-RUBY = #{($nmake && !$extmk && !$configure_args.has_key?('--ruby')) ? '$(ruby:/=\)' : '$(ruby)'}
-RM = $(RUBY) -run -e rm -- -f
-MAKEDIRS = $(RUBY) -run -e mkdir -- -p
-INSTALL_PROG = $(RUBY) -run -e install -- -vpm 0755
-INSTALL_DATA = $(RUBY) -run -e install -- -vpm 0644
+RUBY = $(ruby#{sep})
+RM = #{config_string('RM') || '$(RUBY) -run -e rm -- -f'}
+MAKEDIRS = #{config_string('MAKEDIRS') || '@$(RUBY) -run -e mkdir -- -p'}
+INSTALL = #{config_string('INSTALL') || '@$(RUBY) -run -e install -- -vp'}
+INSTALL_PROG = #{config_string('INSTALL_PROG') || '$(INSTALL) -m 0755'}
+INSTALL_DATA = #{config_string('INSTALL_DATA') || '$(INSTALL) -m 0644'}
+COPY = #{config_string('CP') || '@$(RUBY) -run -e cp -- -v'}
 
 #### End of system configuration section. ####
 
+preload = #{$preload.join(" ") if $preload}
 }
   if $nmake == ?b
     mk.each do |x|
@@ -773,7 +1009,13 @@ INSTALL_DATA = $(RUBY) -run -e install -- -vpm 0644
 end
 
 def dummy_makefile(srcdir)
-  configuration(srcdir) << "all install: Makefile\n" << CLEANINGS
+  configuration(srcdir) << <<RULES << CLEANINGS
+CLEANFILES = #{$cleanfiles.join(' ')}
+DISTCLEANFILES = #{$distcleanfiles.join(' ')}
+
+all install static install-so install-rb: Makefile
+
+RULES
 end
 
 def create_makefile(target, srcprefix = nil)
@@ -798,15 +1040,18 @@ def create_makefile(target, srcprefix = nil)
   srcprefix ||= '$(srcdir)'
   Config::expand(srcdir = srcprefix.dup)
 
-  unless $objs then
+  if not $objs
     $objs = []
-    for f in Dir[File.join(srcdir, "*.{#{SRC_EXT.join(%q{,})}}")]
-      $objs.push(File.basename(f, ".*") << "." << $OBJEXT)
-    end
-  else
-    for i in $objs
-      i.sub!(/\.o\z/, ".#{$OBJEXT}")
+    srcs = Dir[File.join(srcdir, "*.{#{SRC_EXT.join(%q{,})}}")]
+    for f in srcs
+      obj = File.basename(f, ".*") << ".o"
+      $objs.push(obj) unless $objs.index(obj)
     end
+  elsif !(srcs = $srcs)
+    srcs = $objs.collect {|obj| obj.sub(/\.o\z/, '.c')}
+  end
+  for i in $objs
+    i.sub!(/\.o\z/, ".#{$OBJEXT}")
   end
   $objs = $objs.join(" ")
 
@@ -833,80 +1078,107 @@ def create_makefile(target, srcprefix = nil)
   dllib = target ? "$(TARGET).#{CONFIG['DLEXT']}" : ""
   staticlib = target ? "$(TARGET).#$LIBEXT" : ""
   mfile = open("Makefile", "wb")
-  mfile.print configuration(srcdir)
+  mfile.print configuration(srcprefix)
   mfile.print %{
+libpath = #{$LIBPATH.join(" ")}
 LIBPATH = #{libpath}
 DEFFILE = #{deffile}
 
 CLEANFILES = #{$cleanfiles.join(' ')}
 DISTCLEANFILES = #{$distcleanfiles.join(' ')}
 
+extout = #{$extout}
+extout_prefix = #{$extout_prefix}
 target_prefix = #{target_prefix}
 LOCAL_LIBS = #{$LOCAL_LIBS}
 LIBS = #{$LIBRUBYARG} #{$libs} #{$LIBS}
+SRCS = #{srcs.collect(&File.method(:basename)).join(' ')}
 OBJS = #{$objs}
 TARGET = #{target}
 DLLIB = #{dllib}
-STATIC_LIB = #{staticlib}
-}
-  if $extmk
-    mfile.print %{
-RUBYCOMMONDIR = $(rubylibdir)
-RUBYLIBDIR    = $(rubylibdir)$(target_prefix)
-RUBYARCHDIR   = $(archdir)$(target_prefix)
-}
-  else
-    mfile.print %{
-RUBYCOMMONDIR = $(sitedir)$(target_prefix)
-RUBYLIBDIR    = $(sitelibdir)$(target_prefix)
-RUBYARCHDIR   = $(sitearchdir)$(target_prefix)
-}
-  end
-  mfile.print %{
-CLEANLIBS     = "$(TARGET).{lib,exp,il?,tds,map}" $(DLLIB)
-CLEANOBJS     = "*.{#{$OBJEXT},#{$LIBEXT},s[ol],pdb,bak}"
+EXTSTATIC = #{$static || ""}
+STATIC_LIB = #{staticlib unless $static.nil?}
 
-all:		#{target ? "$(DLLIB)" : "Makefile"}
-static:		$(STATIC_LIB)
+}
+  install_dirs.each {|d| mfile.print("%-14s= %s\n" % d) if /^[[:upper:]]/ =~ d[0]}
+  n = ($extout ? '$(RUBYARCHDIR)/' : '') + '$(TARGET).'
+  mfile.print %{
+TARGET_SO     = #{($extout ? '$(RUBYARCHDIR)/' : '')}$(DLLIB)
+CLEANLIBS     = #{n}#{CONFIG['DLEXT']} #{n}il? #{n}tds #{n}map
+CLEANOBJS     = *.#{$OBJEXT} *.#{$LIBEXT} *.s[ol] *.pdb *.exp *.bak
+
+all:		#{$extout ? "install" : target ? "$(DLLIB)" : "Makefile"}
+static:		$(STATIC_LIB)#{$extout ? " install-rb" : ""}
 }
   mfile.print CLEANINGS
   dirs = []
-  if not $static and target
-    dirs << (dir = "$(RUBYARCHDIR)")
-    mfile.print("install: #{dir}\n")
+  mfile.print "install: install-so install-rb\n\n"
+  sodir = (dir = "$(RUBYARCHDIR)").dup
+  mfile.print("install-so: #{dir}\n")
+  if target
     f = "$(DLLIB)"
     dest = "#{dir}/#{f}"
-    mfile.print "install: #{dest}\n"
-    mfile.print "#{dest}: #{f} #{dir}\n\t@$(INSTALL_PROG) #{f} #{dir}\n"
+    mfile.print "install-so: #{dest}\n"
+    unless $extout
+      mfile.print "#{dest}: #{f}\n"
+      if (sep = config_string('BUILD_FILE_SEPARATOR'))
+        f.gsub!("/", sep)
+        dir.gsub!("/", sep)
+        sep = ":/="+sep
+        f.gsub!(/(\$\(\w+)(\))/) {$1+sep+$2}
+        f.gsub!(/(\$\{\w+)(\})/) {$1+sep+$2}
+        dir.gsub!(/(\$\(\w+)(\))/) {$1+sep+$2}
+        dir.gsub!(/(\$\{\w+)(\})/) {$1+sep+$2}
+      end
+      mfile.print "\t$(INSTALL_PROG) #{f} #{dir}\n"
+    end
   end
-  for i in [[["lib/**/*.rb", "$(RUBYLIBDIR)", "lib"]], $INSTALLFILES]
+  mfile.print("install-rb: pre-install-rb install-rb-default\n")
+  mfile.print("install-rb-default: pre-install-rb-default\n")
+  mfile.print("pre-install-rb: Makefile\n")
+  mfile.print("pre-install-rb-default: Makefile\n")
+  for sfx, i in [["-default", [["lib/**/*.rb", "$(RUBYLIBDIR)", "lib"]]], ["", $INSTALLFILES]]
     files = install_files(mfile, i, nil, srcprefix) or next
     for dir, *files in files
       unless dirs.include?(dir)
 	dirs << dir
-	mfile.print("install: #{dir}\n")
+	mfile.print "pre-install-rb#{sfx}: #{dir}\n"
       end
       files.each do |f|
 	dest = "#{dir}/#{File.basename(f)}"
-	mfile.print("install: #{dest}\n")
-	mfile.print("#{dest}: #{f} #{dir}\n\t@$(INSTALL_DATA) #{f} #{dir}\n")
+	mfile.print("install-rb#{sfx}: #{dest}\n")
+	mfile.print("#{dest}: #{f}\n\t$(#{$extout ? 'COPY' : 'INSTALL_DATA'}) ")
+	sep = config_string('BUILD_FILE_SEPARATOR')
+	if sep
+	  f = f.gsub("/", sep)
+	  sep = ":/="+sep
+	  f = f.gsub(/(\$\(\w+)(\))/) {$1+sep+$2}
+	  f = f.gsub(/(\$\{\w+)(\})/) {$1+sep+$2}
+	else
+	  sep = ""
+	end
+	mfile.print("#{f} $(@D#{sep})\n")
       end
     end
   end
-  if dirs.empty?
-    mfile.print("install:\n")
-  else
-    dirs.each {|dir| mfile.print "#{dir}:\n\t@$(MAKEDIRS) #{dir}\n"}
-  end
+  dirs.unshift(sodir) if target and !dirs.include?(sodir)
+  dirs.each {|dir| mfile.print "#{dir}:\n\t$(MAKEDIRS) $@\n"}
 
-  mfile.print "\nsite-install: install\n\n"
+  mfile.print <<-SITEINSTALL
+
+site-install: site-install-so site-install-rb
+site-install-so: install-so
+site-install-rb: install-rb
+
+  SITEINSTALL
 
   return unless target
 
+  mfile.puts SRC_EXT.collect {|ext| ".path.#{ext} = $(VPATH)"} if $nmake == ?b
   mfile.print ".SUFFIXES: .#{SRC_EXT.join(' .')} .#{$OBJEXT}\n"
   mfile.print "\n"
 
-  %w[cc cpp cxx C].each do |ext|
+  CXX_EXT.each do |ext|
     COMPILE_RULES.each do |rule|
       mfile.printf(rule, ext, $OBJEXT)
       mfile.printf("\n\t%s\n\n", COMPILE_CXX)
@@ -919,18 +1191,18 @@ static:		$(STATIC_LIB)
     end
   end
 
-  if makedef
-    mfile.print "$(DLLIB): $(OBJS) $(DEFFILE)\n\t"
-  else
-    mfile.print "$(DLLIB): $(OBJS)\n\t"
-  end
-  mfile.print "@-$(RM) $@\n\t"
-  mfile.print "@-$(RM) $(TARGET).lib\n\t" if $mswin
-  mfile.print LINK_SO, "\n\n"
-  mfile.print "$(STATIC_LIB): $(OBJS)\n\t"
-  mfile.print "$(AR) #{config_string('ARFLAGS') || 'cru '}$@ $(OBJS)"
-  if ranlib = config_string('RANLIB')
-    mfile.print "\n\t@-#{ranlib} $(DLLIB) 2> /dev/null || true"
+  mfile.print "$(RUBYARCHDIR)/" if $extout
+  mfile.print "$(DLLIB): ", (makedef ? "$(DEFFILE) " : ""), "$(OBJS)\n"
+  mfile.print "\t@-$(RM) $@\n"
+  mfile.print "\t@-$(MAKEDIRS) $(@D)\n" if $extout
+  link_so = LINK_SO.gsub(/^/, "\t")
+  mfile.print link_so, "\n\n"
+  unless $static.nil?
+    mfile.print "$(STATIC_LIB): $(OBJS)\n\t"
+    mfile.print "$(AR) #{config_string('ARFLAGS') || 'cru '}$@ $(OBJS)"
+    config_string('RANLIB') do |ranlib|
+      mfile.print "\n\t@-#{ranlib} $(DLLIB) 2> /dev/null || true"
+    end
   end
   mfile.print "\n\n"
   if makedef
@@ -940,28 +1212,81 @@ static:		$(STATIC_LIB)
 
   depend = File.join(srcdir, "depend")
   if File.exist?(depend)
+    suffixes = []
+    depout = []
     open(depend, "r") do |dfile|
       mfile.printf "###\n"
+      cont = implicit = nil
+      impconv = proc do
+	COMPILE_RULES.each {|rule| depout << (rule % implicit[0]) << implicit[1]}
+	implicit = nil
+      end
+      ruleconv = proc do |line|
+	if implicit
+	  if /\A\t/ =~ line
+	    implicit[1] << line
+	    next
+	  else
+	    impconv[]
+	  end
+	end
+	if m = /\A\.(\w+)\.(\w+)(?:\s*:)/.match(line)
+	  suffixes << m[1] << m[2]
+	  implicit = [[m[1], m[2]], [m.post_match]]
+	  next
+	elsif RULE_SUBST and /\A(?!\s*\w+\s*=)[$\w][^#]*:/ =~ line
+	  line.gsub!(%r"(\s)(?!\.)([^$(){}+=:\s\/\\,]+)(?=\s|\z)") {$1 + RULE_SUBST % $2}
+	end
+	depout << line
+      end
       while line = dfile.gets()
 	line.gsub!(/\.o\b/, ".#{$OBJEXT}")
-	line.gsub!(/(\s)([^\s\/]+\.[ch])/, '\1{$(srcdir)}\2') if $nmake
 	line.gsub!(/\$\(hdrdir\)\/config.h/, $config_h) if $config_h
-	mfile.print line
+	if /(?:^|[^\\])(?:\\\\)*\\$/ =~ line
+	  (cont ||= []) << line
+	  next
+	elsif cont
+	  line = (cont << line).join
+	  cont = nil
+	end
+	ruleconv.call(line)
+      end
+      if cont
+	ruleconv.call(cont.join)
+      elsif implicit
+	impconv.call
       end
     end
+    unless suffixes.empty?
+      mfile.print ".SUFFIXES: .", suffixes.uniq.join(" ."), "\n\n"
+    end
+    mfile.print depout
+  else
+    headers = %w[ruby.h defines.h]
+    if RULE_SUBST
+      headers.each {|h| h.sub!(/.*/) {|*m| RULE_SUBST % m}}
+    end
+    headers << $config_h if $config_h
+    headers << "$(RUBY_EXTCONF_H)" if $extconf_h
+    mfile.print "$(OBJS): ", headers.join(' '), "\n"
   end
+
+  $makefile_created = true
 ensure
   mfile.close if mfile
 end
 
 def init_mkmf(config = CONFIG)
+  $makefile_created = false
+  $arg_config = []
   $enable_shared = config['ENABLE_SHARED'] == 'yes'
   $defs = []
+  $extconf_h = nil
   $CFLAGS = with_config("cflags", arg_config("CFLAGS", config["CFLAGS"])).dup
   $ARCH_FLAG = with_config("arch_flag", arg_config("ARCH_FLAG", config["ARCH_FLAG"])).dup
   $CPPFLAGS = with_config("cppflags", arg_config("CPPFLAGS", config["CPPFLAGS"])).dup
   $LDFLAGS = (with_config("ldflags") || "").dup
-  $INCFLAGS = "-I$(topdir)"
+  $INCFLAGS = "-I$(topdir) -I$(hdrdir) -I$(srcdir)"
   $DLDFLAGS = with_config("dldflags", arg_config("DLDFLAGS", config["DLDFLAGS"])).dup
   $LIBEXT = config['LIBEXT'].dup
   $OBJEXT = config["OBJEXT"].dup
@@ -973,6 +1298,7 @@ def init_mkmf(config = CONFIG)
   $INSTALLFILES = nil
 
   $objs = nil
+  $srcs = nil
   $libs = ""
   if $enable_shared or Config.expand(config["LIBRUBY"].dup) != Config.expand(config["LIBRUBY_A"].dup)
     $LIBRUBYARG = config['LIBRUBYARG']
@@ -980,12 +1306,31 @@ def init_mkmf(config = CONFIG)
 
   $LOCAL_LIBS = ""
 
-  $cleanfiles = []
-  $distcleanfiles = []
+  $cleanfiles = config_string('CLEANFILES') {|s| Shellwords.shellwords(s)} || []
+  $distcleanfiles = config_string('DISTCLEANFILES') {|s| Shellwords.shellwords(s)} || []
 
+  $extout ||= nil
+  $extout_prefix ||= nil
+
+  $arg_config.clear
   dir_config("opt")
 end
 
+FailedMessage = <<MESSAGE
+Could not create Makefile due to some reason, probably lack of
+necessary libraries and/or headers.  Check the mkmf.log file for more
+details.  You may need configuration options.
+
+Provided configuration options:
+MESSAGE
+
+def mkmf_failed(path)
+  unless $makefile_created or File.exist?("Makefile")
+    opts = $arg_config.collect {|t, n| "\t#{t}#{"=#{n}" if n}\n"}
+    abort "*** #{path} failed ***\n" + FailedMessage + opts.join
+  end
+end
+
 init_mkmf
 
 $make = with_config("make-prog", ENV["MAKE"] || "make")
@@ -1001,8 +1346,15 @@ end
 Config::CONFIG["srcdir"] = CONFIG["srcdir"] =
   $srcdir = arg_config("--srcdir", File.dirname($0))
 $configure_args["--topsrcdir"] ||= $srcdir
-Config::CONFIG["topdir"] = CONFIG["topdir"] =
-  $curdir = arg_config("--curdir", Dir.pwd)
+if $curdir = arg_config("--curdir")
+  Config.expand(curdir = $curdir.dup)
+else
+  curdir = $curdir = "."
+end
+unless File.expand_path(Config::CONFIG["topdir"]) == File.expand_path(curdir)
+  CONFIG["topdir"] = $curdir
+  Config::CONFIG["topdir"] = curdir
+end
 $configure_args["--topdir"] ||= $curdir
 $ruby = arg_config("--ruby", File.join(Config::CONFIG["bindir"], CONFIG["ruby_install_name"]))
 
@@ -1024,29 +1376,35 @@ COMMON_HEADERS = (hdr.join("\n") unless hdr.empty?)
 COMMON_LIBS = config_string('COMMON_LIBS', &split) || []
 
 COMPILE_RULES = config_string('COMPILE_RULES', &split) || %w[.%s.%s:]
-COMPILE_C = config_string('COMPILE_C') || '$(CC) $(CFLAGS) $(CPPFLAGS) -c $<'
-COMPILE_CXX = config_string('COMPILE_CXX') || '$(CXX) $(CXXFLAGS) $(CPPFLAGS) -c $<'
+RULE_SUBST = config_string('RULE_SUBST')
+COMPILE_C = config_string('COMPILE_C') || '$(CC) $(INCFLAGS) $(CPPFLAGS) $(CFLAGS) -c $<'
+COMPILE_CXX = config_string('COMPILE_CXX') || '$(CXX) $(INCFLAGS) $(CPPFLAGS) $(CXXFLAGS) -c $<'
 TRY_LINK = config_string('TRY_LINK') ||
-  "$(CC) #{OUTFLAG}conftest $(INCFLAGS) -I$(hdrdir) $(CPPFLAGS) " \
+  "$(CC) #{OUTFLAG}conftest $(INCFLAGS) $(CPPFLAGS) " \
   "$(CFLAGS) $(src) $(LIBPATH) $(LDFLAGS) $(ARCH_FLAG) $(LOCAL_LIBS) $(LIBS)"
 LINK_SO = config_string('LINK_SO') ||
   if CONFIG["DLEXT"] == $OBJEXT
-    "ld $(DLDFLAGS) -r -o $(DLLIB) $(OBJS)\n"
+    "ld $(DLDFLAGS) -r -o $@ $(OBJS)\n"
   else
-    "$(LDSHARED) $(DLDFLAGS) $(LIBPATH) #{OUTFLAG}$(DLLIB) " \
+    "$(LDSHARED) $(DLDFLAGS) $(LIBPATH) #{OUTFLAG}$@ " \
     "$(OBJS) $(LOCAL_LIBS) $(LIBS)"
   end
 LIBPATHFLAG = config_string('LIBPATHFLAG') || ' -L"%s"'
 RPATHFLAG = config_string('RPATHFLAG') || ''
 LIBARG = config_string('LIBARG') || '-l%s'
 
+sep = config_string('BUILD_FILE_SEPARATOR') {|sep| ":/=#{sep}" if sep != "/"} || ""
 CLEANINGS = "
 clean:
-		@$(RM) $(CLEANLIBS) $(CLEANOBJS) $(CLEANFILES)
+		@-$(RM) $(CLEANLIBS#{sep}) $(CLEANOBJS#{sep}) $(CLEANFILES#{sep})
 
 distclean:	clean
-		@$(RM) Makefile extconf.h conftest.* mkmf.log
-		@$(RM) core ruby$(EXEEXT) *~ $(DISTCLEANFILES)
+		@-$(RM) Makefile $(RUBY_EXTCONF_H) conftest.* mkmf.log
+		@-$(RM) core ruby$(EXEEXT) *~ $(DISTCLEANFILES#{sep})
 
 realclean:	distclean
 "
+
+if not $extmk and /\A(extconf|makefile).rb\z/ =~ File.basename($0)
+  END {mkmf_failed($0)}
+end
