@@ -3,7 +3,7 @@
   sprintf.c -
 
   $Author: matz $
-  $Date: 2005/09/09 13:15:15 $
+  $Date: 2006/06/12 13:30:32 $
   created at: Fri Oct 15 10:39:26 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -13,6 +13,7 @@
 **********************************************************************/
 
 #include "ruby.h"
+#include "re.h"
 #include <ctype.h>
 #include <math.h>
 
@@ -113,7 +114,11 @@ sign_bits(base, p)
     t = p++; \
     n = 0; \
     for (; p < end && ISDIGIT(*p); p++) { \
-	n = 10 * n + (*p - '0'); \
+	int next_n = 10 * n + (*p - '0'); \
+        if (next_n / 10 != n) {\
+	    rb_raise(rb_eArgError, #val " too big"); \
+	} \
+	n = next_n; \
     } \
     if (p >= end) { \
 	rb_raise(rb_eArgError, "malformed format string - %%*[0-9]"); \
@@ -146,7 +151,7 @@ sign_bits(base, p)
  *
  *    Flag     | Applies to   | Meaning
  *    ---------+--------------+-----------------------------------------
- *    space    | bdeEfgGioxXu | Leave a space at the start of 
+ *    space    | bdeEfgGiouxX | Leave a space at the start of 
  *             |              | positive numbers.
  *    ---------+--------------+-----------------------------------------
  *    (digit)$ | all          | Specifies the absolute argument number
@@ -162,11 +167,11 @@ sign_bits(base, p)
  *             |              | point to be added, even if no digits follow.
  *             |              | For `g' and 'G', do not remove trailing zeros.
  *    ---------+--------------+-----------------------------------------
- *    +        | bdeEfgGioxXu | Add a leading plus sign to positive numbers.
+ *    +        | bdeEfgGiouxX | Add a leading plus sign to positive numbers.
  *    ---------+--------------+-----------------------------------------
  *    -        | all          | Left-justify the result of this conversion.
  *    ---------+--------------+-----------------------------------------
- *    0 (zero) | all          | Pad with zeros, not spaces.
+ *    0 (zero) | bdeEfgGiouxX | Pad with zeros, not spaces.
  *    ---------+--------------+-----------------------------------------
  *    *        | all          | Use the next argument as the field width. 
  *             |              | If negative, left-justify the result. If the
@@ -309,6 +314,10 @@ rb_f_sprintf(argc, argv)
 	  case '5': case '6': case '7': case '8': case '9':
 	    n = 0;
 	    for (; p < end && ISDIGIT(*p); p++) {
+		int next_n = 10 * n + (*p - '0');
+		if (next_n / 10 != n) {
+		    rb_raise(rb_eArgError, "width too big");
+		}
 		n = 10 * n + (*p - '0');
 	    }
 	    if (p >= end) {
@@ -405,6 +414,23 @@ rb_f_sprintf(argc, argv)
 			len = prec;
 		    }
 		}
+		{
+		    char *s, *send;
+		    long l;
+
+			s = RSTRING(str)->ptr;
+			send = s + RSTRING(str)->len;
+			l = 0;
+			while (s < send) {
+			    long n = mbclen(*s);
+			    if (l + n > len) {
+				len = l;
+				break;
+			    }
+			    l += n;
+			    s += n;
+			}
+		}
 		if (flags&FWIDTH) {
 		    if (width > len) {
 			CHECK(width);
@@ -446,6 +472,7 @@ rb_f_sprintf(argc, argv)
 		int base, bignum = 0;
 		int len, pos;
 		VALUE tmp;
+                volatile VALUE tmp1;
 
 		switch (*p) {
 		  case 'd':
@@ -594,7 +621,7 @@ rb_f_sprintf(argc, argv)
 		    val = rb_big_clone(val);
 		    rb_big_2comp(val);
 		}
-		tmp = rb_big2str(val, base);
+		tmp1 = tmp = rb_big2str(val, base);
 		s = RSTRING(tmp)->ptr;
 		if (*s == '-') {
 		    if (base == 10) {
