@@ -2,8 +2,8 @@
 
   time.c -
 
-  $Author: nobu $
-  $Date: 2005/10/22 04:27:48 $
+  $Author: matz $
+  $Date: 2006/07/06 15:34:04 $
   created at: Tue Dec 28 14:31:59 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -139,6 +139,7 @@ time_overflow_p(secp, usecp)
     *usecp = usec;
 }
 
+static VALUE time_new_internal _((VALUE, time_t, time_t));
 static VALUE
 time_new_internal(klass, sec, usec)
     VALUE klass;
@@ -193,7 +194,7 @@ time_timeval(time, interval)
 	    if (f != t.tv_sec) {
 		rb_raise(rb_eRangeError, "%f out of Time range", RFLOAT(time)->value);
 	    }
-	    t.tv_usec = (time_t)(d*1e6);
+	    t.tv_usec = (time_t)(d*1e6+0.5);
 	}
 	break;
 
@@ -486,6 +487,16 @@ tmcmp(a, b)
         return 0;
 }
 
+#if SIZEOF_TIME_T == SIZEOF_LONG
+typedef unsigned long unsigned_time_t;
+#elif SIZEOF_TIME_T == SIZEOF_INT
+typedef unsigned int unsigned_time_t;
+#elif SIZEOF_TIME_T == SIZEOF_LONG_LONG
+typedef unsigned LONG_LONG unsigned_time_t;
+#else
+# error cannot find integer type which size is same as time_t.
+#endif
+
 static time_t
 search_time_t(tptr, utc_p)
     struct tm *tptr;
@@ -499,12 +510,12 @@ search_time_t(tptr, utc_p)
     find_dst = 0 < tptr->tm_isdst;
 
 #ifdef NEGATIVE_TIME_T
-    guess_lo = 1L << (8 * sizeof(time_t) - 1);
+    guess_lo = (time_t)~((unsigned_time_t)~(time_t)0 >> 1);
 #else
     guess_lo = 0;
 #endif
     guess_hi = ((time_t)-1) < ((time_t)0) ?
-               (1UL << (8 * sizeof(time_t) - 1)) - 1 :
+	       (time_t)((unsigned_time_t)~(time_t)0 >> 1) :
 	       ~(time_t)0;
 
     guess = timegm_noleapsecond(tptr);
@@ -1255,16 +1266,6 @@ time_to_s(time)
     return rb_str_new(buf, len);
 }
 
-#if SIZEOF_TIME_T == SIZEOF_LONG
-typedef unsigned long unsigned_time_t;
-#elif SIZEOF_TIME_T == SIZEOF_INT
-typedef unsigned int unsigned_time_t;
-#elif SIZEOF_TIME_T == SIZEOF_LONG_LONG
-typedef unsigned LONG_LONG unsigned_time_t;
-#else
-# error cannot find integer type which size is same as time_t.
-#endif
-
 static VALUE
 time_add(tobj, offset, sign)
     struct time_object *tobj;
@@ -1286,7 +1287,7 @@ time_add(tobj, offset, sign)
     if (f != (double)sec_off)
         rb_raise(rb_eRangeError, "time %s %f out of Time range",
             sign < 0 ? "-" : "+", v);
-    usec_off = (time_t)(d*1e6);
+    usec_off = (time_t)(d*1e6+0.5);
 
     if (sign < 0) {
         sec = tobj->tv.tv_sec - sec_off;
@@ -1899,7 +1900,7 @@ time_mdump(time)
     if ((tm->tm_year & 0xffff) != tm->tm_year)
 	rb_raise(rb_eArgError, "year too big to marshal");
 
-    p = 0x1          << 31 | /*  1 */
+    p = 0x1UL        << 31 | /*  1 */
 	tm->tm_year  << 14 | /* 16 */
 	tm->tm_mon   << 10 | /*  4 */
 	tm->tm_mday  <<  5 | /*  5 */
@@ -1975,12 +1976,12 @@ time_mload(time, str)
 	s |= buf[i]<<(8*(i-4));
     }
 
-    if ((p & (1<<31)) == 0) {
+    if ((p & (1UL<<31)) == 0) {
 	sec = p;
 	usec = s;
     }
     else {
-	p &= ~(1<<31);
+       p &= ~(1UL<<31);
 	tm.tm_year = (p >> 14) & 0xffff;
 	tm.tm_mon  = (p >> 10) & 0xf;
 	tm.tm_mday = (p >>  5) & 0x1f;
