@@ -2,8 +2,8 @@
 
   variable.c -
 
-  $Author: nobu $
-  $Date: 2005/09/28 15:58:32 $
+  $Author: matz $
+  $Date: 2006/04/19 03:01:59 $
   created at: Tue Apr 19 23:55:15 JST 1994
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -469,7 +469,9 @@ mark_global_entry(key, entry)
 void
 rb_gc_mark_global_tbl()
 {
-    st_foreach(rb_global_tbl, mark_global_entry, 0);
+    if (rb_global_tbl) {
+	st_foreach(rb_global_tbl, mark_global_entry, 0);
+    }
 }
 
 static ID
@@ -1668,6 +1670,10 @@ rb_const_set(klass, id, val)
     ID id;
     VALUE val;
 {
+    if (NIL_P(klass)) {
+	rb_raise(rb_eTypeError, "no class/module to define constant %s",
+		 rb_id2name(id));
+    }
     mod_av_set(klass, id, val, Qtrue);
 }
 
